from __future__ import absolute_import
from pycparser import c_ast
from pycparserext import ext_c_parser
from pycparserext.ext_c_generator import OpenCLCGenerator


class Generator(OpenCLCGenerator):
    def visit(self, n):
        result = super(OpenCLCGenerator, self).visit(n)

        # Simple method to remove extra newlines except where required
        # (#pragma). This has the consequence that it'll modify character
        # arrays. Fortunately, character arrays are rarely used in OpenCL in
        # practice.
        lines = [line for line in result.replace("#pragma", "\n#pragma").split("\n") if len(line) > 0]
        result = ""
        for i, line in enumerate(lines):
            if "#pragma" in line:
                # Put pragma on its own line.
                if i > 0 and "#pragma" not in lines[i - 1]:
                    result += "\n"
                result += line + "\n"
            elif len(line) > 0:
                result += line

        return result

    def visit_BinaryOp(self, n):
        # Only include brackets when order cannot be implied through operator
        # priority. Based off of C because I could not find the operator
        # precedence in the specification.
        # http://en.cppreference.com/w/c/language/operator_precedence
        operator_precedence = {
            "*": 3,
            "/": 3,
            "%": 3,
            "+": 4,
            "-": 4,
            "<<": 5,
            ">>": 5,
            "<": 6,
            "<=": 6,
            ">": 6,
            ">=": 6,
            "==": 7,
            "!=": 7,
            "&": 8,
            "^": 9,
            "|": 10,
            "&&": 11,
            "||": 12,
        }
        left = self.visit(n.left)
        right = self.visit(n.right)
        if isinstance(n.left, c_ast.BinaryOp):
            if operator_precedence[n.left.op] > operator_precedence[n.op]:
                left = "(" + left + ")"
        if isinstance(n.right, c_ast.BinaryOp):
            if operator_precedence[n.right.op] >= operator_precedence[n.op]:
                right = "(" + right + ")"
        result = "%s%s%s" % (left, n.op, right)
        return result.replace(" ", "")

    def visit_Assignment(self, n):
        result = self.visit(n.lvalue)
        result += n.op
        result += self._parenthesize_if(n.rvalue, lambda n: isinstance(n, c_ast.Assignment))
        return result

    def visit_Decl(self, n, no_type=False):
        result = n.name
        if not no_type:
            result = self._generate_decl(n)
        if n.bitsize:
            result += ":" + self.visit(n.bitsize)
        if n.init:
            result += "=" + self._visit_expr(n.init)
        return result

    def visit_DeclList(self, n):
        result = ""
        for index, decl in enumerate(n.decls):
            if index == 0:
                result = self.visit(decl)
            else:
                result += "," + self.visit_Decl(decl, no_type=True)
        return result

    def visit_Cast(self, n):
        # Work around an issue (bug?) I noticed in green's compiler that causes
        # vector literals in double brackets to be treated as a scaler using the
        # last value in the expression list.
        # ushort4 test = (ushort4)((0,1,2,3)); //test = (3,3,3,3)
        # ushort4 test = (ushort4)(0,1,2,3); //test = (0,1,2,3)
        result = "(" + self._generate_type(n.to_type) + ")"
        if self._is_simple_node(n.expr):
            result += " " + self.visit(n.expr)
        else:
            result += "(" + self.visit(n.expr) + ")"
        return result.replace(") ", ")")

    def visit_UnaryOp(self, n):
        # Avoid extra brackets around the expression unless required by the
        # operator. Just sizeof() needs the extra brackets.
        result = self.visit(n.expr)
        if n.op == "sizeof":
            result = "sizeof(" + result + ")"
        # Add postfix operators at the end.
        elif len(n.op) >= 1 and n.op[0] == "p":
            result = result + n.op[1:]
        else:
            result = n.op + result
        return result

    def visit_TernaryOp(self, n):
        result = self.visit(n.cond)
        result += "?" + self.visit(n.iftrue)
        result += ":" + self.visit(n.iffalse)
        return result

    def visit_If(self, n):
        result = "if("
        if n.cond:
            result += self.visit(n.cond)
        result += ")"
        result += self._generate_stmt(n.iftrue).strip()
        if n.iffalse:
            result += "else"
            if isinstance(n.iffalse, c_ast.If):
                result += " "
            elif not self._is_multi_stmt_compound(n.iffalse):
                result += " "
            result += self._generate_stmt(n.iffalse).strip()
        return result

    def visit_For(self, n):
        result = "for("
        if n.init:
            result += self.visit(n.init)
        result += ";"
        if n.cond:
            result += self.visit(n.cond)
        result += ";"
        if n.next:
            result += self.visit(n.next)
        result += ")"
        result += self._generate_stmt(n.stmt).strip()
        return result

    def visit_While(self, n):
        result = "while("
        if n.cond:
            result += self.visit(n.cond)
        result += ")"
        result += self._generate_stmt(n.stmt).strip()
        return result

    def visit_DoWhile(self, n):
        result = "do"
        if not self._is_multi_stmt_compound(n.stmt):
            result += " "
        result += self._generate_stmt(n.stmt).strip()
        result += "while("
        if n.cond:
            result += self.visit(n.cond)
        result += ");"
        return result

    def visit_Struct(self, n):
        result = "struct"
        if n.name:
            result += " " + n.name
        if n.decls:
            result += "{" + self._generate_grouped_stmts(n.decls) + "}"
        return result

    def visit_Switch(self, n):
        result = "switch(" + self.visit(n.cond) + ")"
        result += self._generate_stmt(n.stmt)
        return result

    def visit_Enum(self, n):
        result = "enum"
        if n.name:
            result += " " + n.name
        if n.values:
            result += "{"
            for index, enum in enumerate(n.values.enumerators):
                result += enum.name
                if enum.value:
                    result += "=" + self.visit(enum.value)
                if index != len(n.values.enumerators) - 1:
                    result += ","
            result += "}"
        return result

    def visit_Case(self, n):
        result = "case " + self.visit(n.expr) + ":"
        for stmt in n.stmts:
            result += self._generate_stmt(stmt).strip()
        return result

    def visit_Default(self, n):
        result = "default:"
        for stmt in n.stmts:
            result += self._generate_stmt(stmt).strip()
        return result

    def visit_FuncDef(self, n):
        result = self.visit(n.decl)

        # Remove double spaces created by pycparserext when the __kernel and
        # __attribute__ modifiers are both specified.
        result = result.replace("  ", " ")

        # Remove extra space after declaring an attribute modifier.
        result = result.replace(")) ", "))")

        body = self.visit(n.body)
        if self._is_multi_stmt_compound(n.body):
            result += body
        else:
            result += "{" + body + "}"
        return result

    def visit_ParamList(self, n):
        return ",".join(self.visit(param) for param in n.params)

    def visit_ExprList(self, n):
        return ",".join([self.visit(expr) for expr in n.exprs])

    def visit_InitList(self, n):
        return ",".join([self.visit(expr) for expr in n.exprs])

    def visit_Compound(self, n):
        if not n.block_items:
            return "{}"
        result = self._generate_grouped_stmts(n.block_items)
        if len(n.block_items) > 1:
            result = "{" + result + "}"
        return result

    def visit_FileAST(self, n):
        # Prevent parent implementation from inserting an unnecessary newline
        # for non-function definitions at the top level.
        result = ""
        for ext in n.ext:
            if isinstance(ext, c_ast.FuncDef) or isinstance(ext, c_ast.Pragma):
                result += self.visit(ext)
            else:
                result += self.visit(ext) + ";"
        return result

    def _generate_type(self, n, modifiers=[]):
        result = super(Generator, self)._generate_type(n, modifiers)

        # Remove space between closing bracket and name.
        if "declname" in n.attr_names and n.declname and result.endswith("} " + n.declname):
            space_index = result.rfind("} ")
            result = result[:space_index + 1] + result[space_index + 2:]

        # Remove space around pointer indicator.
        result = result.replace("* ", "*").replace(" *", "*")

        return result

    def _generate_grouped_stmts(self, stmts):
        """Generate a list of statements into a string. Sequences of
           declarations with the same type are grouped into compact form. For
           example, "float a = 0; float b; float c = 1.0f;" becomes
           "float a=0,b,c=1.0f;".
        """

        # Format a sequence of declarations in compact form.
        def group_declarations(decl_type, start_index, end_index):
            result = " ".join(decl_type)

            for x in range(start_index, end_index):
                is_ptr = isinstance(stmts[x].type, c_ast.PtrDecl)
                is_struct_decl = isinstance(stmts[x].type.type, c_ast.Struct) and stmts[x].type.type.decls
                is_enum_values = isinstance(stmts[x].type.type, c_ast.Enum) and stmts[x].type.type.values
                if x == start_index and not is_ptr and not is_struct_decl and not is_enum_values:
                    result += " "
                if x != start_index:
                    result += ","
                decl = stmts[x]
                if is_ptr:
                    result += "*" + decl.type.type.declname
                else:
                    result += decl.type.declname
                if isinstance(decl.type, ext_c_parser.TypeDeclExt):
                    result += " __attribute__((" + self.visit(decl.type.attributes) + "))"
                if decl.init:
                    result += "=" + self._generate_stmt(decl.init)[:-2]  # Strip ";\n" at end.

            result += ";"
            return result

        # Format each statement while searching for sequences of declarations.
        result = ""
        decl_chain_start_index = -1
        decl_chain_type = []
        for i, stmt in enumerate(stmts):
            # Type declaration statement.
            if isinstance(stmt, c_ast.Decl) and (isinstance(stmt.type, c_ast.TypeDecl) or isinstance(stmt.type, c_ast.PtrDecl)):
                typedecl = stmt.type.type if isinstance(stmt.type, c_ast.TypeDecl) else stmt.type.type.type
                if isinstance(typedecl, c_ast.Struct):
                    decl_type = [self.visit_Struct(typedecl), ]
                elif isinstance(typedecl, c_ast.Enum):
                    decl_type = [self.visit_Enum(typedecl), ]
                else:
                    decl_type = typedecl.names

                # New sequence and no sequence currently found.
                if decl_chain_start_index == -1:
                    decl_chain_start_index = i
                    decl_chain_type = decl_type
                # New sequence but a sequence was already found. Process it
                # first.
                elif decl_type != decl_chain_type:
                    result += group_declarations(decl_chain_type, decl_chain_start_index, i)
                    decl_chain_start_index = i
                    decl_chain_type = decl_type
            # Typical statement.
            else:
                # Format existing sequence before generating next statement.
                if decl_chain_start_index != -1:
                    result += group_declarations(decl_chain_type, decl_chain_start_index, i)
                    decl_chain_start_index = -1
                result += self._generate_stmt(stmt)

        # Format final sequence if list of statements ends in a sequence.
        if decl_chain_start_index != -1:
            result += group_declarations(decl_chain_type, decl_chain_start_index, len(stmts))

        return result

    def _is_multi_stmt_compound(self, n):
        return isinstance(n, c_ast.Compound) and ((n.block_items and len(n.block_items) != 1) or not n.block_items)



from __future__ import absolute_import
from pycparser.c_lexer import CLexer
from pycparserext import ext_c_lexer


class OpenCLCLexer(ext_c_lexer.OpenCLCLexer):
    pass

# Override pycparserext's implementation because it fails to parse
# some pragmas like #pragma unroll. The original implementation works
# just fine for our minification-related usage.
OpenCLCLexer.t_PPHASH = CLexer.t_PPHASH

# Add support for __const type qualifier. It's not in any of the specs but it
# is commonly used in red's OpenCL examples.
ext_c_lexer.add_lexer_keywords(OpenCLCLexer, ["__const"])



from __future__ import absolute_import
import sys
import unittest
sys.path.insert(0, "..")
from oclminify.build import try_build
from oclminify.minify import minify


class TestMinifier(unittest.TestCase):
    def assert_minify(self, data, expected_result, **kwargs):
        if not try_build(data):
            raise AssertionError("Input OpenCL code could not be built.")
        result = minify(data, **kwargs)
        self.assertEqual(result, expected_result)
        if not try_build(result):
            raise AssertionError("Minified OpenCL code could not be built.")

    def test_simple(self):
        data = r"""
            __kernel void main()
            {
            }"""
        self.assert_minify(data, "__kernel void a(){}")

    def test_no_extra_casting_brackets(self):
        data = r"""
            __kernel void main()
            {
                ushort4 test = (ushort4)(0,1,2,3);
                ushort test2 = (ushort)1.0f;
            }"""
        self.assert_minify(data, "__kernel void a(){ushort4 b=(ushort4)(0,1,2,3);ushort c=(ushort)1.0f;}")
        data = r"""
            __kernel void main()
            {
                float4 test1 = (float4)(0.0f,1.0f,2.0f,3.0f);
                ushort4 test2 = convert_ushort4(test1 - (float4)(0.0f,1.0f,2.0f,3.0f));
                test2 = convert_ushort4((float4)(0.0f,1.0f,2.0f,3.0f) - (float4)(0.0f,1.0f,2.0f,3.0f));
            }"""
        self.assert_minify(data, "__kernel void a(){float4 b=(float4)(0.0f,1.0f,2.0f,3.0f);ushort4 c=convert_ushort4(b-(float4)(0.0f,1.0f,2.0f,3.0f));c=convert_ushort4((float4)(0.0f,1.0f,2.0f,3.0f)-(float4)(0.0f,1.0f,2.0f,3.0f));}")

    def test_simple_operator_precedence(self):
        data = r"""
            __kernel void main()
            {
                float test1 = 1.0f;
                int test2 = 2;
                float test3 = test1 / (2.0f * (float)test2);
            }"""
        self.assert_minify(data, "__kernel void a(){float b=1.0f;int c=2;float d=b/(2.0f*(float)c);}")

    def test_no_brackets_when_implied_operator_precedence(self):
        data = r"""
            __kernel void main()
            {
                int a = 3;
                int b = 5;
                int test = a * b - a + b / a;
                test = a * (b - a) + b / a;
                test = (b - a) * (b - a);
            }"""
        self.assert_minify(data, "__kernel void a(){int b=3,c=5,d=b*c-b+c/b;d=b*(c-b)+c/b;d=(c-b)*(c-b);}")
        data = r"""
        __kernel void main()
        {
            int a = 3;
            int b = 5;
            int test = (((((((((((((((((a % b) / a) * b) - a) + b) >> a) << b) >= a) > b) <= a) < b) != a) == b) & a) ^ b) | a) && b) || a;
        }"""
        self.assert_minify(data, "__kernel void a(){int b=3,c=5,d=b%c/b*c-b+c>>b<<c>=b>c<=b<c!=b==c&b^c|b&&c||b;}")

    def test_brackets_when_no_operator_precedence(self):
        data = r"""
        __kernel void main()
        {
            int a = 3;
            int b = 5;
            int test = (((((((((((((((((a || b) && a) | b) ^ a) & b) == a) != b) < a) <= b) > a) >= b) << a) >> b) + a) - b) * a) / b) % a;
        }"""
        self.assert_minify(data, "__kernel void a(){int b=3,c=5,d=(((((((((b||c)&&b)|c)^b)&c)==b!=c)<b<=c>b>=c)<<b>>c)+b-c)*b/c%b;}")

    def test_shrink_data_types(self):
        data = r"""
            __kernel void main()
            {
                typedef unsigned int t;
                t test = 0;
            }"""
        self.assert_minify(data, "__kernel void a(){typedef uint b;b c=0;}")
        data = r"""
            typedef unsigned int typedef_uint;
            __kernel void main()
            {
                unsigned char a = 0;
                unsigned short b = 0;
                unsigned int c = 0;
                unsigned long d = 0;
                typedef_uint e = 0;
            }"""
        self.assert_minify(data, "typedef uint a;__kernel void b(){uchar c=0;ushort d=0;uint e=0;ulong f=0;a g=0;}")
        data = r"""
            __kernel void main()
            {
                int test = sizeof(unsigned char);
            }"""
        self.assert_minify(data, "__kernel void a(){int b=sizeof(uchar);}")
        data = r"""
            __kernel void main()
            {
                uint* test = 0;
                *(volatile unsigned int*)test;
            }"""
        self.assert_minify(data, "__kernel void a(){uint*b=0;*(volatile uint*)b;}")
        data = r"""
            unsigned int func(unsigned int arg)
            {
            }"""
        self.assert_minify(data, "uint a(uint b){}")

    def test_preserve_attributes(self):
        data = r"""
            typedef unsigned int aligned_uint __attribute__ ((alligned(8)));
            __kernel void main()
            {
                aligned_uint test = 0;
                int test2[8] __attribute__((unused));
            }"""
        self.assert_minify(data, "typedef uint a __attribute__((alligned(8)));__kernel void b(){a c=0;int d[8] __attribute__((unused));}")

    def test_struct(self):
        data = r"""
            typedef unsigned int typedef_uint;
            struct TestStruct
            {
                unsigned char test1;
                unsigned short test2;
                unsigned int test3;
                unsigned long test4;
                typedef_uint test5;
                float4 test6;
                unsigned int test7 __attribute__ ((alligned(8)));
            };
            __kernel void main()
            {
                struct TestStruct ts;
                ts.test1 = 0;
                ts.test2 = 0;
                ts.test3 = 0;
                ts.test4 = 0;
                ts.test5 = 0;
                ts.test6.xy = (float2)0.0f;
                ts.test7 = 0;
            }"""
        self.assert_minify(data, "typedef uint a;struct b{uchar a;ushort b;uint c;ulong d;a e;float4 f;uint g __attribute__((alligned(8)));};__kernel void c(){struct b d;d.a=0;d.b=0;d.c=0;d.d=0;d.e=0;d.f.xy=(float2)0.0f;d.g=0;}")
        data = r"""
            __kernel void main()
            {
                struct
                {
                    unsigned char test;
                } ts;
                ts.test = 0;
            }"""
        self.assert_minify(data, "__kernel void a(){struct{uchar a;}b;b.a=0;}")
        data = r"""
            __kernel void main()
            {
                typedef struct Node
                {
                    struct Node* node;
                } node;
            }"""
        self.assert_minify(data, "__kernel void a(){typedef struct b{struct b*a;}c;}")

    def test_struct_inside_struct(self):
        data = r"""
            struct Struct1
            {
                unsigned short value;
            };
            struct Struct2
            {
                unsigned short value;
                struct Struct1 struct1;
                struct Struct3
                {
                    unsigned short value;
                    struct Struct1 struct1;
                };
                struct Struct3 struct3;
                struct
                {
                    unsigned short value;
                    struct Struct1 struct1;
                } struct4;
            };
            __kernel void main()
            {
                struct Struct2 test;
                test.value = 0;
                test.struct1.value = 0;
                test.struct3.value = 0;
                test.struct3.struct1.value = 0;
                test.struct4.value = 0;
                test.struct4.struct1.value = 0;
            }"""
        self.assert_minify(data, "struct a{ushort a;};struct b{ushort a;struct a b;struct c{ushort a;struct a b;};struct c d;struct{ushort a;struct a b;}e;};__kernel void c(){struct b d;d.a=0;d.b.a=0;d.d.a=0;d.d.b.a=0;d.e.a=0;d.e.b.a=0;}")

    def test_struct_rvalue_reference(self):
        # User specified function.
        data = r"""
            struct TestStruct
            {
                unsigned short value1;
                float8 value2;
            };
            struct TestStruct Func()
            {
                struct TestStruct result;
                result.value1 = 0;
                result.value2 = (float8)0.0f;
                return result;
            }
            __kernel void main()
            {
                unsigned short value1 = Func().value1;
                float4 value2 = Func().value2.s0123;
            }"""
        self.assert_minify(data, "struct a{ushort a;float8 b;};struct a b(){struct a c;c.a=0;c.b=(float8)0.0f;return c;}__kernel void c(){ushort d=b().a;float4 e=b().b.lo;}")
        # Built-in functions.
        data = r"""
            __kernel void main()
            {
                int8 test = (int8)0;
                float4 test2 = convert_float8(test).s0123;
                test2 = as_float8(test).s0123;
            }"""
        self.assert_minify(data, "__kernel void a(){int8 b=(int8)0;float4 c=convert_float8(b).lo;c=as_float8(b).lo;}")

    @unittest.expectedFailure  # TODO: Support figuring out the return type of generic built-in functions based off input argument type(s).
    def test_struct_rvalue_reference_fail(self):
        data = r"""
            __kernel void main()
            {
                float8 test = (float8)0.0f;
                sin(test).s0123;
                sin(test + (float8)1.0f).s0123;
                sin(test + convert_float8((int8)0)).s0123;
            }"""
        self.assert_minify(data, "__kernel void a(){float8 b=(float8)0.0f;sin(b).lo;sin(b+(float8)1.0f).lo;sin(b+convert_float8((int8)0)).lo;}")

    def test_shrink_vector_indices(self):
        data = r"""
            __kernel void main()
            {
                float4 test = (float4)(0.0f,1.0f,2.0f,3.0f);
                test = test.s0123; //.xyzw

                uchar16 test2 = (uchar16)(1);
                uchar8 test3 = test2.s01234567; //.lo
                test3 = test2.s89abcdef; //.hi
                test3 = test2.s02468ace; //.even
                test3 = test2.s13579bdf; //.odd

                float2 test4 = test.even; //.xz
                test4 = test.odd; //.yw
            }"""
        self.assert_minify(data, "__kernel void a(){float4 b=(float4)(0.0f,1.0f,2.0f,3.0f);b=b.xyzw;uchar16 c=(uchar16)1;uchar8 d=c.lo;d=c.hi;d=c.even;d=c.odd;float2 e=b.xz;e=b.yw;}")

    def test_enum(self):
        data = r"""
            enum
            {
                ANON_UNVALUED,
                ANON_VALUED = 100
            };
            enum Named
            {
                NAMED_UNVALUED,
                NAMED_VALUED = 100
            };
            __kernel void main()
            {
                int anonVar = ANON_UNVALUED;
                anonVar = ANON_VALUED;

                enum Named namedVar = NAMED_UNVALUED;
                namedVar = NAMED_VALUED;

                enum
                {
                    ANONVAR_UNVALUED,
                    ANONVAR_VALUED = 100
                } enumVar;
                enumVar = ANONVAR_UNVALUED;
                enumVar = ANONVAR_VALUED;
            }"""
        self.assert_minify(data, "enum{a,b=100};enum c{d,e=100};__kernel void f(){int g=a;g=b;enum c h=d;h=e;enum{i,j=100}k;k=i;k=j;}")

    def test_do_while(self):
        data = r"""
            __kernel void main()
            {
                int test = 0;
                do
                {
                    test += 1;
                } while(test < 5);
            }"""
        self.assert_minify(data, "__kernel void a(){int b=0;do b+=1;while(b<5);}")

    def test_switch(self):
        data = r"""
            __kernel void main()
            {
                int value = 0;
                switch(value)
                {
                    case 0:
                        value = 5;
                        break;
                    default:
                        value = 4;
                        break;
                };
            }"""
        self.assert_minify(data, "__kernel void a(){int b=0;switch(b){case 0:b=5;break;default:b=4;break;};}")

    def test_ternary(self):
        data = r"""
            __kernel void main()
            {
                int test1 = 5;
                int test2 = test1 < 6 ? 1 + 1 : 10 * 10;
            }"""
        self.assert_minify(data, "__kernel void a(){int b=5,c=b<6?1+1:10*10;}")

    def test_init_list(self):
        data = r"""
            __kernel void main()
            {
                float test[4] = {
                    0.0f,
                    1.0f,
                    2.0f,
                    3.0f
                };
            }"""
        self.assert_minify(data, "__kernel void a(){float b[4]={0.0f,1.0f,2.0f,3.0f};}")

    def test_preserve_pragmas(self):
        data = r"""
            #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            __kernel void main()
            {
            }"""
        self.assert_minify(data, "#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n__kernel void a(){}")
        data = r"""
            __kernel void main()
            {
                #pragma unroll
                for(uint x = 0;x < 16;x++)
                {
                }
            }"""
        self.assert_minify(data, "__kernel void a(){\n#pragma unroll\nfor(uint b=0;b<16;b++){}}")

    def test_statement_brackets(self):
        # Single statement, no brackets.
        data = r"""
            __kernel void main()
            {
                int test = 0;
                if(true)
                {
                    test = 0;
                }
                else if(false)
                {
                    test = 0;
                }
                else
                {
                    test = 0;
                }

                while(true)
                {
                    test = 0;
                }

                do
                {
                    test = 0;
                } while(true);

                for(;;)
                {
                    test = 0;
                }
            }"""
        self.assert_minify(data, "__kernel void a(){int b=0;if(true)b=0;else if(false)b=0;else b=0;while(true)b=0;do b=0;while(true);for(;;)b=0;}")
        # Multiple statements, brackets.
        data = r"""
            __kernel void main()
            {
                int test = 0;
                if(true)
                {
                    test = 0;
                    test = 1;
                }
                else if(false)
                {
                    test = 0;
                    test = 1;
                }
                else
                {
                    test = 0;
                    test = 1;
                }

                while(true)
                {
                    test = 0;
                    test = 1;
                }

                do
                {
                    test = 0;
                    test = 1;
                } while(true);

                for(;;)
                {
                    test = 0;
                    test = 1;
                }
            }"""
        self.assert_minify(data, "__kernel void a(){int b=0;if(true){b=0;b=1;}else if(false){b=0;b=1;}else{b=0;b=1;}while(true){b=0;b=1;}do{b=0;b=1;}while(true);for(;;){b=0;b=1;}}")

    def test_group_variable_declarations(self):
        data = r"""
            __kernel void main()
            {
                struct Test
                {
                    float first;
                    float second;
                    float third;
                    int fourth;
                    float fifth;
                };

                float first;
                float second = 1.0f;
                float third;
                int fourth;
                float fifth;
            }"""
        self.assert_minify(data, "__kernel void a(){struct b{float a,b,c;int d;float e;};float c,d=1.0f,e;int f;float g;}")
        data = r"""
            __kernel void main()
            {
                struct Test
                {
                    float* first;
                    float* second;
                    float* third;
                    int* fourth;
                    float* fifth;
                };

                float* first;
                float* second = 0;
                float* third;
                int* fourth;
                float* fifth;
            }"""
        self.assert_minify(data, "__kernel void a(){struct b{float*a,*b,*c;int*d;float*e;};float*c,*d=0,*e;int*f;float*g;}")

    def test_minify_kernel_names(self):
        data = r"""
            void func()
            {
            }
            __kernel void main()
            {
            }"""
        self.assert_minify(data, "void a(){}__kernel void b(){}", minify_kernel_names=True)
        self.assert_minify(data, "void a(){}__kernel void main(){}", minify_kernel_names=False)

    def test_global_postfix(self):
        data = r"""
            typedef uint gtypedef;
            struct gstruct
            {
                gtypedef var;
            };
            void gfunc()
            {
            }
            __kernel void gkernel()
            {
                typedef uint ltypedef;
                struct lstruct
                {
                    gtypedef gvar;
                    ltypedef lvar;
                };
            }"""
        self.assert_minify(data, "typedef uint a_global;struct b_global{a_global a;};void c_global(){}__kernel void d_global(){typedef uint a;struct b{a_global a;a b;};}", global_postfix="_global")

    def test_function_args(self):
        data = r"""
            void func(__global float* arg0,int arg1,float arg2[10])
            {
            }"""
        self.assert_minify(data, "void a(__global float*b,int c,float d[10]){}")

    @unittest.expectedFailure
    def test_remove_unncessary_vector_indices(self):
        # TODO: See if this can be implemented easily. This will probably break some of the other tests.
        data = r"""
            __kernel void main()
            {
                float4 test = (float4)(0.0f,1.0f,2.0f,3.0f);
                float4 test2 = test.xyzw;
                test2 = test.s0123;
            }"""
        self.assert_minify(data, "__kernel void a(){float4 a=(float4)(0.0f,1.0f,2.0f,3.0f),b=a;b=a;")


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
from __future__ import print_function
import os
import sys


def try_build(data):
    """Try to build OpenCL script to test for errors instead of having to wait
    until after the minified version is used during run time which is much
    harder to debug.
    """

    if isinstance(data, str) and sys.version_info.major >= 3:
        data = data.encode("utf-8", "ignore")

    try:
        # PyOpenCL is required.
        import pyopencl as cl
        import pyopencl.cffi_cl as _cl

        # Use the default context and the first device because it's
        # convenient. We could add a command line flag later if necessary.
        ctx = cl.create_some_context(interactive=False)
        device = ctx.devices[0]

        # Just build and check for errors. Some drivers output whitespace in
        # the log, so  we strip for consistency to prevent printing a blank
        # line to stdout.
        prg = None
        try:
            os.environ["PYOPENCL_NO_CACHE"] = "true"
            prg = cl.Program(ctx, data).build(devices=[device, ])
        except _cl.RuntimeError as e:
            # Print error log but strip off the parts PyOpenCL adds since we
            # don't need them.
            print("\n".join(e.what.split("\n")[0:-2]), file=sys.stderr)
            print("Could not build program.", file=sys.stderr)
            return False
        log = prg.get_build_info(device, cl.program_build_info.LOG).strip()
        if len(log) > 0:
            print(log, file=sys.stderr)
        if prg.get_build_info(device, cl.program_build_info.STATUS) != 0:
            print("Error while building program.", file=sys.stderr)
            return False
    except ImportError:
        print("PyOpenCL not found. Skipping test program build.", file=sys.stderr)
    except IndexError:
        print("No OpenCL devices detected. Skipping test program build.", file=sys.stderr)
    return True



from __future__ import absolute_import
from pycparser import CParser
from pycparserext import ext_c_parser


class Parser(ext_c_parser.OpenCLCParser):
    # Allow extension specific types to be parsed. Even those in newer versions
    # of OpenCL that we don't really support should parse when possible.
    initial_type_symbols = ext_c_parser.OpenCLCParser.initial_type_symbols | \
                           set(["void",  # OpenCL >= 1.0
                                "counter32_t",  # cl_ext_atomic_counters_32
                                "counter64_t",  # cl_ext_atomic_counters_64
                                "atomic_int", ])  # TODO: Find out which standard introduced this. OpenCL 2.0, I think.

    # Use our patched lexer instead. See lexer.py for details.
    from oclminify.lexer import OpenCLCLexer as lexer_class

    # Add support for __const type qualifier. It's not in any of the specs but
    # it is commonly used in red's OpenCL examples.
    def p_type_qualifier_cl(self, p):
        """ type_qualifier  : __GLOBAL
                            | GLOBAL
                            | __LOCAL
                            | LOCAL
                            | __CONSTANT
                            | __CONST
                            | CONSTANT
                            | __PRIVATE
                            | PRIVATE
                            | __READ_ONLY
                            | READ_ONLY
                            | __WRITE_ONLY
                            | WRITE_ONLY
                            | __READ_WRITE
                            | READ_WRITE
        """
        p[0] = p[1]

# Override pycparserext's implementation because it fails to parse pointers in
# type declarations. Once again, the original implementation works just fine
# for our minification-related usage.
Parser.p_declarator_2 = CParser.p_declarator_2



VECTOR_SIZES = ["", "2", "3", "4", "8", "16"]
ROUNDING_MODES = ["", "_rte", "_rtz", "_rtp", "_rtn"]


# Internal only helper functions
def _join_dicts(*dicts):
    joined_dicts = {}
    for dict in dicts:
        joined_dicts.update(dict)
    return joined_dicts


def _vector_size(arg_type):
    return "".join([char for char in arg_type if char.isdigit()])


def _vector_type(arg_type):
    return "".join([char for char in arg_type if char.isalpha()])


def _strip_vector_size_from_first_arg(func_name, args):
    assert(len(args) >= 1)
    return _vector_type(args[0])


def _abs_function_return_type(func_name, args):
    assert(len(args) >= 1)
    vector_type = args[0]
    if not vector_type.startswith("u"):
        vector_type = "u" + vector_type
    return vector_type


def _upsample_function_return_type(func_name, args):
    assert(len(args) >= 1)
    vector_size = _vector_size(args[0])
    vector_type = _vector_type(args[0])
    unsigned = "u" if vector_type.startswith("u") else ""
    if unsigned:
        vector_type = vector_type[1:]

    return unsigned + {
        "char": "short",
        "short": "int",
        "int": "long",
    }[vector_type] + vector_size


def _relational_function_return_type(func_name, args):
    assert(len(args) >= 1)
    arg0 = args[0]
    vector_size = _vector_size(arg0)
    if arg0.startswith("float"):
        return "int" + vector_size
    else:
        return "long" + vector_size


def _shuffle_function_return_type(func_name, args):
    assert(len(args) >= 2)
    return _vector_type(args[0]) + _vector_size(args[-1])


class BUILTIN:
    CONSTANTS = {
        "HUGE_VAL": "double",
        "HUGE_VALF": "float",
        "INFINITY": "float",
        "M_E": "double",
        "M_E_F": "float",
        "M_LOG2E": "double",
        "M_LOG2E_F": "float",
        "M_LOG10E": "double",
        "M_LOG10E_F": "float",
        "M_LN2": "double",
        "M_LN2_F": "float",
        "M_LN10": "double",
        "M_LN10_F": "float",
        "M_PI": "double",
        "M_PI_F": "float",
        "M_PI_2": "double",
        "M_PI_2_F": "float",
        "M_PI_4": "double",
        "M_PI_4_F": "float",
        "M_1_PI": "double",
        "M_1_PI_F": "float",
        "M_2_PI": "double",
        "M_2_PI_F": "float",
        "M_2_SQRTPI": "double",
        "M_2_SQRTPI_F": "float",
        "M_SQRT2": "double",
        "M_SQRT2_F": "float",
        "M_SQRT1_2": "double",
        "M_SQRT1_2_F": "float",
        "MAXFLOAT": "float",
        "NAN": "float",
    }

    # All convert_ and as_ functions that specify their return type as part of
    # their name.
    CAST_FUNCTIONS = ["convert_%s%s" % (type, size)
                        for type in ["char", "uchar", "short", "ushort", "int", "uint", "long", "ulong"]
                            for size in VECTOR_SIZES] + \
                     ["convert_%s%s%s%s" % (type, size, sat, rounding)
                        for type in ["float", "double"]
                            for size in VECTOR_SIZES
                                for sat in ["", "_sat"]
                                    for rounding in ROUNDING_MODES] + \
                     ["as_%s%s" % (type, size)
                        for type in ["char", "uchar", "short", "ushort", "int", "uint", "long", "ulong", "float", "double"]
                            for size in VECTOR_SIZES]

    # Functions who's return type is the same as their first argument's type.
    GEN1_FUNCTIONS = [
        # Math Functions (float/double)
        "acos",
        "acosh",
        "acospi"
        "asin",
        "asinh",
        "asinpi",
        "atan",
        "atan2",
        "atanh",
        "atanpi",
        "atan2pi",
        "cbrt",
        "ceil",
        "copysign",
        "cos",
        "cosh",
        "cospi",
        "erfc",
        "erf",
        "exp",
        "exp2",
        "exp10",
        "expm1",
        "fabs",
        "fdim",
        "floor",
        "fma",
        "fmax",
        "fmin",
        "fmod",
        "fract",
        "frexp",
        "hypot",
        "ldexp",
        "lgamma",
        "lgamma_r",
        "log",
        "log2",
        "log10",
        "log1p",
        "logb",
        "mad",
        "maxmag",
        "minmag",
        "modf",
        "nextafter",
        "pow"
        "pown",
        "powr",
        "remainder",
        "remquo",
        "rint",
        "rootn",
        "round",
        "rsqrt",
        "sin",
        "sincos",
        "sinh",
        "sinpi",
        "sqrt",
        "tan",
        "tanh",
        "tanpi",
        "tgamma",
        "trunc",
        "half_cos",
        "half_divide",
        "half_exp",
        "half_exp2",
        "half_exp10",
        "half_log",
        "half_log2",
        "half_log10",
        "half_powr",
        "half_recip",
        "half_rsqrt",
        "half_sin",
        "half_sqrt",
        "half_tan",
        "native_cos",
        "native_divide",
        "native_exp",
        "native_exp2",
        "native_exp10",
        "native_log",
        "native_log2",
        "native_log10",
        "native_powr",
        "native_recip",
        "native_rsqrt",
        "native_sin",
        "native_sqrt",
        "native_tan",
        # Math Functions (integer)
        "add_sat",
        "hadd",
        "rhadd",
        "clamp",
        "clz",
        "mad_hi",
        "mad_sat",
        "max",
        "min",
        "mul_hi",
        "rotate",
        "sub_sat",
        "popcount",
        "mad24",
        "mul24",
        # Common Functions
        # "clamp", # Already listed as an integer math function above.
        "degrees",
        # "max", # Already listed as an integer math function above.
        # "min", # Already listed as an integer math function above.
        "mix",
        "radians",
        "step",
        "smoothstep",
        "sign",
        # Geometric Functions
        "cross",
        "normalize"
        "fast_normalize",
        # Relational Functions
        "bitselect",
        "select",
        # Atomic Functions
        "atomic_inc",
        "atomic_dec",
    ]

    # Functions who's return type is the same as their second argument's type.
    GEN2_FUNCTIONS = [
        # Atomic Functions
        "atomic_add",
        "atomic_sub",
        "atomic_xchg",
        "atomic_cmpxchg",
        "atomic_min",
        "atomic_max",
        "atomic_and",
        "atomic_or",
        "atomic_xor",
    ]

    # Functions that always have the same return type. The return type is
    # specified as the value here.
    FIXED_FUNCTIONS_MAP = _join_dicts({
        # Work-Item Functions
        "get_work_dim": "uint",
        "get_global_size": "size_t",
        "get_global_id": "uint",
        "get_local_size": "size_t",
        "get_local_id": "size_t",
        "get_num_groups": "size_t",
        "get_group_id": "size_t",
        "get_global_offset": "size_t",
        # Relational Functions
        "any": "int",
        "all": "int", },
        # Vector Data Load and Store Functions
        dict([("vstore%s" % size, "void") for size in VECTOR_SIZES]),
        dict([("vload_half%s" % size, "float%s" % size) for size in VECTOR_SIZES]),
        dict([("vstore_half%s%s" % (size, rounding), "void") for size in VECTOR_SIZES for rounding in ROUNDING_MODES]),
        dict([("vstorea_half%s%s" % (size, rounding), "void") for size in VECTOR_SIZES[1:] for rounding in ROUNDING_MODES]),
        dict([("vloada_half%s" % size, "float%s" % size) for size in VECTOR_SIZES[1:]]), {
        # Synchronization Functions
        "barrier": "void",
        # Explicit Memory Fence Functions
        "mem_fence": "void",
        "read_mem_fence": "void",
        "write_mem_fence": "void",
        # Async Copies from Global to Local Memory, Local to Global Memory, and
        # Prefetch Functions
        "async_work_group_copy": "event_t",
        "async_work_group_strided_copy": "event_t",
        "wait_group_events": "void",
        "prefetch": "void",
        # Miscellaneous Vector Functions
        "vec_step": "int",
        # printf Function (It has its own section in the documentation...)
        "printf": "int",
        # Image Read and Write Functions
        "read_imagef": "float4",
        "read_imagei": "int4",
        "read_imageui": "uint4",
        "write_imagef": "void",
        "write_imagei": "void",
        "write_imageui": "void",
        # Built-in Image Query Functions
        "get_image_width": "int",
        "get_image_height": "int",
        "get_image_depth": "int",
        "get_image_channel_data_type": "int",
        "get_image_channel_order": "int",
        "get_image_array_size": "size_t",
    })

    # Functions that require different processing from the above choices to
    # determine the return type. The values in the map are functions that take
    # the function name and a list of function argument types.
    OTHER_FUNCTIONS_MAP = _join_dicts({
        # Math Functions (float/double)
        "ilogb": lambda func_name, args: "int" + _vector_size(args[0]),
        "nan": lambda func_name, args: ("float%s" if args[0][:4] == "uint" else "double%s") % args[0][4:],
        # Math Functions (integer)
        "abs": _abs_function_return_type,
        "abs_diff": _abs_function_return_type,
        "upsample": _upsample_function_return_type,
        # Geometric Functions
        "dot": _strip_vector_size_from_first_arg,
        "distance": _strip_vector_size_from_first_arg,
        "length": _strip_vector_size_from_first_arg,
        "fast_distance": _strip_vector_size_from_first_arg,
        "fast_length": _strip_vector_size_from_first_arg,
        # Relational Functions
        "isequal": _relational_function_return_type,
        "isnotequal": _relational_function_return_type,
        "isgreater": _relational_function_return_type,
        "isgreaterequal": _relational_function_return_type,
        "isless": _relational_function_return_type,
        "islessequal": _relational_function_return_type,
        "islessgreater": _relational_function_return_type,
        "isfinite": _relational_function_return_type,
        "isinf": _relational_function_return_type,
        "isnan": _relational_function_return_type,
        "isnormal": _relational_function_return_type,
        "isordered": _relational_function_return_type,
        "isunordered": _relational_function_return_type,
        "signbit": _relational_function_return_type, },
        # Vector Data Load and Store Functions
        dict([("vload%s" % (size), lambda func_name, args: _vector_type(args[1]) + _vector_size(func_name)) for size in VECTOR_SIZES]), {
        # Miscellaneous Vector Functions
        "shuffle": _shuffle_function_return_type,
        "shuffle2": _shuffle_function_return_type,
        # Built-in Image Query Functions
        "get_image_dim": lambda func_name, args: "int3" if args[0] == "image3d_t" else "int2",
    })

    @classmethod
    def get_func_return_type(cls, func_name, args):
        # Return type is in function name.
        if func_name in cls.CAST_FUNCTIONS:
            # Extract return type from name.
            return func_name.split("_")[1]

        # Return type matches first argument.
        if func_name in cls.GEN1_FUNCTIONS and len(args) >= 1:
            return args[0]

        # Return type matches second argument.
        if func_name in cls.GEN2_FUNCTIONS and len(args) >= 2:
            return args[1]

        # Return type never changes and can be determined solely by the
        # function name.
        if func_name in cls.FIXED_FUNCTIONS_MAP.keys():
            return cls.FIXED_FUNCTIONS_MAP[func_name]

        # Return type requires special handling to figure out.
        if func_name in cls.OTHER_FUNCTIONS_MAP .keys():
            return cls.OTHER_FUNCTIONS_MAP[func_name](func_name, args)

        # Unknown built-in function.
        return "void"



#!/bin/python
from __future__ import absolute_import
from __future__ import print_function
import argparse
from io import open
import os
import sys
import zlib
from oclminify.minify import DEFAULT_PREPROCESSOR_COMMAND, _do_minify
from oclminify.build import try_build


def main():
    # Compatibility check to make sure an old version of pycparser was not
    # installed by accident. This won't be required once a version of pycparser
    # newer than 2.14 is released.
    from pycparser import c_ast
    if "Pragma" not in dir(c_ast):
        print("Your version of pycparser is too old.\n#pragma declarations"
              "will cause minification to fail!\nRun \"pip uninstall "
              "pycparser && pip install "
              "git+https://github.com/eliben/pycparser.git@ffd8cb7dfc4b80c79a500e27736db8f7bfc1186e#egg=pycparser-2.14\"",
              file=sys.stderr)

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="Version 0.6.0\nMinify OpenCL source files.",
                                     epilog="OpenCL is a trademark of Apple Inc., used under license by Khronos.\nCopyright (c) 2016 StarByte Software, Inc. All rights reserved.")
    parser.add_argument("--preprocessor-command", type=str, default=DEFAULT_PREPROCESSOR_COMMAND, help="Command to preprocess input source before minification. Defaults to \"%s\"" % DEFAULT_PREPROCESSOR_COMMAND)
    parser.add_argument("--no-preprocess", action="store_true", default=False, help="Skip preprocessing step. Implies --no-minify.")
    parser.add_argument("--no-minify", action="store_true", default=False, help="Skip minification step. Useful when debugging.")
    parser.add_argument("--compress", action="store_true", default=False, help="Compress output using zlib.")
    parser.add_argument("--strip-zlib-header", action="store_true", default=False, help="Strips the two byte zlib header from the compressed output when --compress is used.")
    parser.add_argument("--header", action="store_true", default=False, help="Embed output in a C header file.")
    parser.add_argument("--header-function-args", action="store_true", default=False, help="Include function argument mappings in C header file.")
    parser.add_argument("--minify-kernel-names", action="store_true", default=False, help="Replace kernel function names with shorter names.")
    parser.add_argument("--global-postfix", type=str, default="", help="Postfix appended to each symbol name in the global scope. Used for preventing name collisions when minifying multiple source files separately. Implies --minify-kernel-names.")
    parser.add_argument("--try-build", action="store_true", default=False, help="Try to build the input using an OpenCL compiler before minifying. The compiled output is discarded. Requires pyopencl.")
    parser.add_argument("--output-file", type=str, default="", help="File path where output should be saved. Omit to write to stdout.")
    parser.add_argument("input", help="File path to OpenCL file that should be minified. A \"-\" indicates that input should be read from stdin.")
    args = parser.parse_args()

    # Read input from the specified file. If the specified file is "-", just
    # read from stdin so text can be piped in from a shell or whatever.
    if args.input == "-":
        data = ""
        while True:
            c = sys.stdin.read(1)
            if len(c) == 0:
                break
            data += c
    else:
        # Read entire file into memory.
        fd = open(args.input, "rb")
        if not fd:
            print("Could not open input file.", file=sys.stderr)
            sys.exit(-1)
        data = fd.read()
        fd.close()

    if args.try_build:
        if not try_build(data):
            sys.exit(-1)

    # Perform preprocessing and minification.
    original_size = len(data)
    original_data = data
    minifier, data = _do_minify(data,
                                preprocessor_command=args.preprocessor_command,
                                minify=not args.no_minify,
                                minify_kernel_names=(args.minify_kernel_names or len(args.global_postfix) > 0) and not args.no_minify,
                                global_postfix=args.global_postfix)
    if args.no_preprocess:
        data = original_data
    minified_size = len(data)

    # Perform zlib compression.
    if not isinstance(data, bytes):
        data = data.encode("utf-8", "ignore")
    if args.compress:
        compressed_data = zlib.compress(data, 9)
        if args.strip_zlib_header:
            # Strip header: 0x78 0xDA
            data = compressed_data[2:]
        else:
            data = compressed_data

    # Print sizes of output after each stage. This is a minifier, might as well
    # see how well we did.
    compressed_size = len(data)
    result_message = "Original Size: %i, Minified Size: %i" % (original_size, minified_size)
    if args.compress:
        result_message += ", Compressed Size: %i" % compressed_size
    print(result_message, file=sys.stderr)

    # Transform minified output into a C header file if run with --header.
    if args.header:
        guard_name = os.path.split(args.input)[-1].upper().replace(".", "_") + "_DATA_H"
        var_base_name = os.path.split(args.input)[-1].lower()
        var_base_name = var_base_name[:var_base_name.find(".")].capitalize()
        header_text = "#ifndef %s\n#define %s\n\n" % (guard_name, guard_name)
        header_text += "static const size_t %s_SIZE = %i;\n" % (var_base_name.upper(), len(data))
        header_text += "static const unsigned char %s_DATA[] = {" % var_base_name.upper()
        for (i, byte) in enumerate(data):
            if isinstance(byte, str):
                byte = ord(byte)
            header_text += str(hex(byte))
            if i != (len(data) - 1):
                header_text += ","
        header_text += "};\n\n"
        for (old_name, func) in minifier.functions.items():
            if old_name not in minifier.kernel_functions:
                continue  # Not a kernel, no need to make available.
            header_text += "#define %s_FUNCTION_%s \"%s\"\n" % (var_base_name.upper(), old_name.upper(), func.name)
            if args.header_function_args:
                for (old_arg, new_arg) in minifier.functions_args[old_name].items():
                    header_text += "#define %s_FUNCTION_%s_ARG_%s = \"%s\"\n" % (var_base_name.upper(), old_name.upper(), old_arg.upper(), new_arg)
                header_text += "\n"
        if not args.header_function_args:
            header_text += "\n"
        header_text += "#endif"
        data = header_text

    # Save output to file if run with --output-file, otherwise just print to
    # stdout so it can be processed further in a shell or whatever.
    if args.output_file == "":
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        print(data)
    else:
        fd = open(args.output_file, "wb")
        if not fd:
            print("Could not open output file", file=sys.stderr)
            sys.exit(-1)
        if not isinstance(data, bytes):
            data = data.encode("utf-8", "ignore")
        fd.write(data)
        fd.close()

if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import print_function
import subprocess
import sys
from oclminify.generator import Generator
from oclminify.minifier import Minifier
from oclminify.parser import Parser


DEFAULT_PREPROCESSOR_COMMAND = "gcc -E -undef -P -std=c99 -"
DEFAULT_MINIFY = True
DEFAULT_MINIFY_KERNEL_NAMES = True
DEFAULT_GLOBAL_POSTFIX = ""


def _do_minify(data,
               preprocessor_command=DEFAULT_PREPROCESSOR_COMMAND,
               minify=DEFAULT_MINIFY,
               minify_kernel_names=DEFAULT_MINIFY_KERNEL_NAMES,
               global_postfix=DEFAULT_GLOBAL_POSTFIX):
    if isinstance(data, str) and sys.version_info.major >= 3:
        data = data.encode("utf-8", "ignore")

    # Use GCC to do the preprocessing.
    p = subprocess.Popen(preprocessor_command.split(" "),
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    data, err = p.communicate(data)
    if err:
        print(err, file=sys.stderr)
    if p.returncode != 0:
        print("Failed to preprocess file", file=sys.stderr)
        sys.exit(-1)
    data = data.decode("utf-8")
    data = data.replace("\r", "")  # Strip Windows newline character added by GCC on Windows.
    preprocessed_data = data

    parser = Parser()
    ast = parser.parse(data)

    # Uncomment when debugging to show the parsed graph.
    # ast.show()

    # Walk source tree graph and apply minification. The minifier is run even
    # when not minifying so we can collect kernel names for header output.
    minifier = Minifier(minify_kernel_names, global_postfix)
    minifier.visit(ast)
    if minify:
        data = Generator().visit(ast)
    else:
        data = preprocessed_data
    return (minifier, data)


def minify(data,
           preprocessor_command=DEFAULT_PREPROCESSOR_COMMAND,
           minify=DEFAULT_MINIFY,
           minify_kernel_names=DEFAULT_MINIFY_KERNEL_NAMES,
           global_postfix=DEFAULT_GLOBAL_POSTFIX):
    return _do_minify(data,
                      preprocessor_command=preprocessor_command,
                      minify=minify,
                      minify_kernel_names=minify_kernel_names,
                      global_postfix=global_postfix)[1]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import copy
import itertools
import sys
from pycparser import c_ast
from oclminify.parser import Parser


class Minifier(c_ast.NodeVisitor):
    IGNORE_TYPE_SYMBOLS = Parser.initial_type_symbols | set(["char", "int", "short", "long", "float", "double"])
    CONSTANT_SYMBOLS = [
        "true",
        "false",
        "CLK_LOCAL_MEM_FENCE",
        "CLK_GLOBAL_MEM_FENCE",
        "CLK_NORMALIZED_COORDS_TRUE",
        "CLK_NORMALIZED_COORDS_FALSE",
        "CLK_ADDRESS_MIRRORED_REPEAT",
        "CLK_ADDRESS_REPEAT",
        "CLK_ADDRESS_CLAMP_TO_EDGE",
        "CLK_ADDRESS_CLAMP",
        "CLK_ADdRESS_NONE",
        "CLK_FILTER_NEAREST",
        "CLK_FILTER_LINEAR",
        "CLK_SNORM_INT8",
        "CLK_SNORM_INT16",
        "CLK_UNORM_INT8",
        "CLK_UNORM_INT16",
        "CLK_UNORM_SHORT_565",
        "CLK_UNORM_SHORT_555",
        "CLK_UNORM_SHORT_101010",
        "CLK_SIGNED_INT8",
        "CLK_SIGNED_INT16",
        "CLK_SIGNED_INT32",
        "CLK_UNSIGNED_INT8",
        "CLK_UNSIGNED_INT16",
        "CLK_UNSIGNED_INT32",
        "CLK_HALF_FLOAT",
        "CLK_FLOAT",
        "CLK_A",
        "CLK_R",
        "CLK_Rx",
        "CLK_RG",
        "CLK_RGx",
        "CLK_RA",
        "CLK_RGB",
        "CLK_RGBx",
        "CLK_RGBA",
        "CLK_ARGB",
        "CLK_BGRA",
        "CLK_INTENSITY",
        "CLK_LUMINANCE",
        # OpenCL >= 2.0
        "CLK_IMAE_MEM_FENCE",
        "memory_order_relaxed",
        "memory_order_acquire",
        "memory_order_release",
        "memory_order_acq_rel",
        "memory_order_seq_cst",
        "memory_scope_work_item",
        "memory_scope_work_group",
        "memory_scope_device",
        "memory_scope_all_svm_devices",
    ]

    class Function(object):
        def __init__(self, name="", return_type=None):
            self.name = name
            self.return_type = return_type

    class Declaration(object):
        def __init__(self):
            self.name = ""
            self.type = ""
            self.children = {}
            self.is_definition = False

        def __eq__(self, other):
            if isinstance(other, Minifier.Declaration):
                return self.name == other.name
            elif type(other) is str:
                return self.name == other
            else:
                return NotImplemented

        def __repr__(self):
            return "(%s) %s %s" % (self.type, self.name, repr(self.children))

    def __init__(self, replace_kernel_names, global_postfix):
        self.functions = {}
        self.functions_args = {}
        self.kernel_functions = []
        self.declaration_scopes = [{}]
        self.replace_kernel_names = replace_kernel_names
        self.global_postfix = global_postfix

    def generic_visit(self, node):
        if node is None:
            return
        else:
            print("FOUND UNSUPPORTED NODE: " + str(node), file=sys.stderr)
            print("Could not find method Minifier.visit_%s" % node.__class__.__name__, file=sys.stderr)
            node.show()
            sys.exit(-1)

    def visit_Constant(self, node):
        pass  # Unused.

    def visit_ID(self, node):
        node.name = self._get_new_declaration_name(node.name)

    def visit_ArrayRef(self, node):
        self.visit(node.name)
        self.visit(node.subscript)

    def visit_StructRef(self, node):
        # Get a list of sequential child struct references so we can walk them
        # and shorten any names as appropriate.
        refs = []
        node_ref = node
        while isinstance(node_ref, c_ast.StructRef):
            refs.append(node_ref)
            node_ref = node_ref.name

        # Walk through struct references in outer most order first
        # (ie. first.second.third). This way we can check the type of the
        # current reference when it's a member of another struct.
        last_ref_type = []
        for node_ref in reversed(refs):
            if not isinstance(node_ref.name, c_ast.StructRef):
                self.visit(node_ref.name)
            decl_type = self._get_structref_type(node_ref)
            if decl_type is None:
                decl_type = last_ref_type
            last_ref_type = []

            # If accessing an anonymous struct, try and use the shortened IDs.
            if isinstance(decl_type, Minifier.Declaration):
                declaration = decl_type
                if node_ref.field.name in declaration.children:
                    child = declaration.children[node_ref.field.name]
                    node_ref.field.name = child.name
                    last_ref_type = child.type
            elif len(decl_type) == 1:
                # If accessing indices of a vector, try and shorten the syntax.
                vector_components = "".join([char for char in decl_type[0] if char.isdigit()])
                vector_types = ["".join(a) for a in itertools.product(["char", "uchar", "short", "ushort", "int", "uint", "long", "ulong", "float", "double", "half"], ["2", "3", "4", "8", "16"])]
                if decl_type[0] in vector_types:
                    node_ref.field.name = self._shorten_vector_access(int(vector_components), node_ref.field.name)
                # If accessing a struct, try and use the shortened IDs.
                if decl_type[0] not in self.IGNORE_TYPE_SYMBOLS:
                    declaration = self._get_declaration_by_new_name(decl_type[0])
                    if declaration and declaration.type == "struct":
                        if node_ref.field.name in declaration.children:
                            last_ref_type = declaration.children[node_ref.field.name].type
                            if not isinstance(last_ref_type, list) and not isinstance(last_ref_type, Minifier.Declaration):
                                last_ref_type = [last_ref_type, ]
                            node_ref.field.name = declaration.children[node_ref.field.name].name

    def visit_FuncCall(self, node):
        node.name.name = self._get_new_function_name(node.name.name)
        self.visit(node.args)

    def visit_UnaryOp(self, node):
        self.visit(node.expr)

    def visit_BinaryOp(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Assignment(self, node):
        self.visit(node.rvalue)
        self.visit(node.lvalue)

    def visit_IdentifierType(self, node):
        # Use shorter idenitifiers when available.
        if node.names == ["unsigned", "char"]:
            node.names = ["uchar"]
        elif node.names == ["unsigned", "short"]:
            node.names = ["ushort"]
        elif node.names == ["unsigned", "int"]:
            node.names = ["uint"]
        elif node.names == ["unsigned", "long"]:
            node.names = ["ulong"]
        if len(node.names) == 1:
            # If the identifier is another built in, skip.
            if node.names[0] in self.IGNORE_TYPE_SYMBOLS:
                return
            # If the identifier is a typedef, use the shortened name.
            new_name = self._get_new_declaration_name(node.names[0])
            if new_name and new_name != node.names[0]:
                node.names = [new_name, ]

    def visit_Decl(self, node, no_type=False):
        if node.init:
            self.visit(node.init)
        if node.funcspec and "__kernel" in node.funcspec:
            self.kernel_functions.append(node.name)
        self.visit(node.type)

    def visit_DeclList(self, node):
        for decl in node.decls:
            self.visit(decl)

    def visit_Cast(self, node):
        self.visit(node.to_type)
        if node.expr:
            self.visit(node.expr)

    def visit_ExprList(self, node):
        for expr in node.exprs:
            self.visit(expr)

    def visit_FileAST(self, node):
        for ext in node.ext:
            self.visit(ext)

    def visit_Compound(self, node):
        self.declaration_scopes.append({})
        if node.block_items is not None:
            for item in node.block_items:
                self.visit(item)
        self.declaration_scopes.pop()

    def visit_ParamList(self, node):
        for param in node.params:
            self.visit(param)

    def visit_Return(self, node):
        if node.expr:
            self.visit(node.expr)

    def visit_Continue(self, node):
        pass  # Unused.

    def visit_Break(self, node):
        pass  # Unused.

    def visit_If(self, node):
        if node.cond:
            self.visit(node.cond)
        self.visit(node.iftrue)
        if node.iffalse:
            self.visit(node.iffalse)

    def visit_For(self, node):
        if node.init:
            self.visit(node.init)
        if node.cond:
            self.visit(node.cond)
        if node.next:
            self.visit(node.next)
        self.visit(node.stmt)

    def visit_FuncDecl(self, node):
        # This should be only reachable for functions declared without
        # definitions (Function Prototype).
        # TODO: Check if the function has a definition provided and make sure
        # they share the same name.
        pass

    def visit_FuncDef(self, node):
        if "__kernel" in node.decl.funcspec:
            self.kernel_functions.append(node.decl.name)

        # Shorten function name unless it's a kernel and Minifier() was
        # initialized with replace_kernel_names=False. If outputting to a
        # C header, you probably want to replace kernel names. In other
        # situations, you probably don't want to replace kernel names or you
        # won't know which kernel is which.
        old_name = node.decl.type.type.declname
        if "__kernel" in node.decl.funcspec and not self.replace_kernel_names:
            new_name = old_name
        else:
            new_name = self._generate_unique_declaration_name()
        node.decl.type.type.declname = new_name

        self.functions[old_name] = Minifier.Function(new_name, [])  # Reserve the function name.
        self.declaration_scopes.append({})
        self.visit(node.decl.type.type.type)  # Return type
        self.visit(node.decl.type.args)  # Args
        self.functions_args[old_name] = self.declaration_scopes[-1]
        self.functions[old_name] = Minifier.Function(new_name, node.decl.type.type.type)  # Include return type after it's processed.

        self.visit(node.body)
        self.declaration_scopes.pop()

    def visit_PtrDecl(self, node):
        self.visit(node.type)

    def visit_ArrayDecl(self, node):
        self.visit(node.type)

    def visit_TypeDecl(self, node):
        self.visit(node.type)
        new_name = self._generate_unique_declaration_name()
        decl = Minifier.Declaration()
        decl.name = new_name

        if "names" in node.type.attr_names:
            decl.type = node.type.names
        elif isinstance(node.type, c_ast.Struct) and node.type.name is None:
            decl.type = self._struct_to_declaration(node.type)
        elif isinstance(node.type, c_ast.Struct):
            decl.type = self._get_declaration_by_new_name(node.type.name, "struct")

        if decl.type is None:
            if "name" in node.type.attr_names:
                decl.type = [node.type.name, ]
            else:
                print("DEBUG: No types for %s" % node.declname, file=sys.stderr)

        self.declaration_scopes[-1][node.declname] = decl
        node.declname = new_name

    def visit_TernaryOp(self, node):
        self.visit(node.cond)
        self.visit(node.iftrue)
        self.visit(node.iffalse)

    def visit_While(self, node):
        self.visit(node.cond)
        self.visit(node.stmt)

    def visit_Typedef(self, node):
        self.visit(node.type)

    def visit_TypeDeclExt(self, node):
        self.visit_TypeDecl(node)

    def visit_Struct(self, node):
        # If an anonymous struct, ignore because it requires special handling
        # to shrink the type declaration names and still make them accessible
        # later.
        if not node.name:
            return
        # Named struct being declared. Shorten name of struct and map the names
        # of its variables to something shorter.
        if node.decls:
            decl = self._struct_to_declaration(node)
            old_name = node.name
            node.name = decl.name
            self.declaration_scopes[-1][old_name] = decl
        # Struct is being declared as type of a variable, just get the new
        # name.
        else:
            node.name = self._get_new_declaration_name(node.name)

    def visit_Enum(self, node):
        if node.name:
            # Shrink name if provided. Considering the required enum prefix when
            # delcaring a variable of this type, we could probably have a set of
            # symbols in the current scope dedicated to enums but it's probably not
            # worth the added complexity.
            if node.values:
                old_name = node.name
                node.name = self._generate_unique_declaration_name()
                self.declaration_scopes[-1][old_name] = node
            # Enum is being declared as a type of a variable, just get the
            # new name.
            else:
                node.name = self._get_new_declaration_name(node.name)

        # Shrink names of each enum symbol.
        if node.values:
            for enum in node.values.enumerators:
                old_name = enum.name
                enum.name = self._generate_unique_declaration_name()
                self.declaration_scopes[-1][old_name] = enum

    def visit_EmptyStatement(self, node):
        pass  # Unused.

    def visit_InitList(self, node):
        for expr in node.exprs:
            self.visit(expr)

    def visit_Pragma(self, node):
        pass  # Unused.

    def visit_Typename(self, node):
        while "type" in dir(node) and not isinstance(node, c_ast.IdentifierType):
            node = node.type
        if isinstance(node, c_ast.IdentifierType):
            self.visit_IdentifierType(node)

    def visit_DoWhile(self, node):
        self.visit(node.stmt)
        self.visit(node.cond)

    def visit_Switch(self, node):
        self.visit(node.cond)
        self.visit(node.stmt)

    def visit_Case(self, node):
        self.visit(node.expr)
        for stmt in node.stmts:
            self.visit(stmt)

    def visit_Default(self, node):
        for stmt in node.stmts:
            self.visit(stmt)

    # TODO: Support these other node types where applicable to OpenCL.
    # def visit_Label(self, node):
    # def visit_Goto(self, node):
    # def visit_EllipsisParam(self, node):
    # def visit_Union(self, node):
    # def visit_NamedInitializer(self, node):

    def _index_to_alpha_str(self, index):
        # TODO: Make this alpha numeric for the first character only and alpha numeric + other supported symbols for the rest.
        if index == 0:
            return 'a'
        characters = list(range(ord('a'), ord('z') + 1)) + list(range(ord('A'), ord('Z') + 1))
        result = ""
        digit = 1
        while index >= 0:
            digit_min_value = len(characters) ** (digit - 1)
            if index < digit_min_value:
                break
            value = index % ((len(characters) ** digit) // len(characters) ** (digit - 1))
            result += chr(characters[value])
            index -= value
            digit += 1
        return "".join(reversed(result))

    def _get_new_function_name(self, name):
        if name in self.functions:
            return self.functions[name].name
        # Remaining functions are probably built-in.
        return name

    def _get_function_by_new_name(self, new_name):
        for func in self.functions.values():
            if func.name == new_name:
                return func
        # Remaining functions are probably built-in.
        return None

    def _is_declaration_name_unique(self, name):
        for scope in self.declaration_scopes:
            if name in [declaration.name for declaration in scope.values()]:
                return False
        return True

    def _generate_unique_declaration_name(self):
        postfix = self.global_postfix if len(self.declaration_scopes) == 1 else ""

        index = 0
        while True:
            name = self._index_to_alpha_str(index) + postfix
            index += 1
            # Make sure declaration name is not currently in use within the
            # visible scope. Also make sure the declaration name does not shadow
            # an existing declaration name.
            if self._is_declaration_name_unique(name) and not self._get_function_by_new_name(name):
                return name

    def _get_new_declaration_name(self, name):
        if name in self.CONSTANT_SYMBOLS:
            return name

        for scope in reversed(self.declaration_scopes):
            if name in scope:
                return scope[name].name
        print("Could not find new declaration name for '%s'" % name, file=sys.stderr)
        return name

    def _get_declaration_by_name(self, name, type_filters=None):
        def is_type_in_filters(decl_type):
            if type_filters is None:
                return True
            for type_filter in type_filters:
                if isinstance(decl_type, str) and isinstance(type_filter, str):
                    if decl_type == type_filter:
                        return True
                elif isinstance(type_filter, type) and isinstance(decl_type, type_filter):
                    return True
            return False

        for scope in reversed(self.declaration_scopes):
            if name in scope:
                if is_type_in_filters(scope[name].type):
                    return copy.deepcopy(scope[name])
        print("Could not find new declaration for '%s'" % name, file=sys.stderr)

    def _get_declaration_by_new_name(self, new_name, type_filter=None):
        for scope in reversed(self.declaration_scopes):
            for declaration in scope.values():
                if declaration.name == new_name:
                    if type_filter is None or declaration.type == type_filter:
                        return copy.deepcopy(declaration)
        print("Could not find declaration with new name '%s'" % new_name, file=sys.stderr)
        return None

    def _get_structref_type(self, node):
        def get_builtin_func_type(func_name, args):
            # Extract type of each argument so we can determine the return type
            # of the built-in function. Many, if not most, built-in functions
            # determine their return type based off of their arguments. If the
            # argument type is a pointer, the pointer part is stripped because
            # it makes it easier to figure out the return type. None of the
            # built-in functions, as of OpenCL 1.2, return a pointer type.
            arg_types = []
            for arg in args.exprs:
                if isinstance(arg, c_ast.ID):
                    declaration = self._get_declaration_by_new_name(arg.name)
                    if declaration is not None:
                        arg_types.append(declaration.type[0])
                        continue
                # TODO: Support figuring out the type of an expression.
                arg_types.append("void")

            from oclminify.functions import BUILTIN
            result = BUILTIN.get_func_return_type(func_name, arg_types)
            return result

        # Find type of struct variable being referenced. It might be accessed
        # as the result of a function, after indexing into an array, or just
        # through a variable directly. Structs inside of structs are a special
        # case handled by the caller of this function (visit_StructRef()).
        if isinstance(node.name, c_ast.StructRef):
            return None
        elif isinstance(node.name, c_ast.FuncCall):
            func_name = node.name.name.name
            func = self._get_function_by_new_name(func_name)
            if func is None:
                return [get_builtin_func_type(func_name, node.name.args), ]
            new_name = func.return_type.name
        elif isinstance(node.name, c_ast.UnaryOp):
            new_name = node.name.expr.name
        elif isinstance(node.name, c_ast.ArrayRef):
            new_name = node.name.name.name
        else:
            new_name = node.name.name

        declaration = self._get_declaration_by_new_name(new_name)
        if declaration:
            if declaration.type == "struct":
                decl_type = declaration.name
            else:
                decl_type = declaration.type
        else:
            decl_type = ["", ]

        return decl_type

    def _struct_to_declaration(self, node):
        declaration = Minifier.Declaration()
        declaration.type = "struct"
        declaration.is_definition = True

        # Generate a short name for this struct.
        declaration.name = self._generate_unique_declaration_name()

        # Generate short names for each declaration in the struct.
        self.declaration_scopes.append({node.name: declaration})
        self.declaration_scopes.append({})  # Allow outer struct name, defined immediately above, to be shadowed.
        for index, node_decl in enumerate(node.decls):
            if isinstance(node_decl.type, c_ast.Struct):
                minify_decl = self._struct_to_declaration(node_decl.type)
                old_declname = node_decl.type.name

                # Replace the generated name with a likely shorter one because
                # this is the scope inside of a struct. We don't have to worry
                # about shadowing.
                new_declname = self._index_to_alpha_str(index)
                minify_decl.name = new_declname
                node_decl.type.name = new_declname

                self.declaration_scopes[-1][old_declname] = minify_decl
            else:
                if isinstance(node_decl.type, c_ast.PtrDecl):
                    node_decl = node_decl.type

                minify_decl = Minifier.Declaration()
                old_declname = node_decl.type.declname
                self.visit(node_decl.type)

                # Generate child declaration name. It just has to be unique to this
                # struct, not to the scope or anything like that.
                new_declname = self._index_to_alpha_str(index)
                node_decl.type.declname = new_declname

                # Save as a child declaration so can rewrite references to this
                # declaration to use the shortened name.
                if isinstance(node_decl.type.type, c_ast.Struct):
                    minify_decl = self._get_declaration_by_name(old_declname, ["struct", Minifier.Declaration])
                    minify_decl.name = node_decl.type.declname
                else:
                    minify_decl.name = node_decl.type.declname
                    minify_decl.type = node_decl.type.type.names[0]

                declaration.children[old_declname] = minify_decl
        self.declaration_scopes.pop()
        self.declaration_scopes.pop()

        return declaration

    def _shorten_vector_access(cls, vector_component_count, vector_components_specified):
        # For vectors of 1-4 components in the form .s0123, convert to the form .xyzw.
        # According to the standard, the .xyzw syntax is only valid for vectors with
        # no more than 4 components.
        # https://www.khronos.org/registry/cl/specs/opencl-1.1.pdf#page=165
        if vector_component_count <= 4 and vector_components_specified.startswith("s"):
            return "".join(map(lambda a: {"0": "x", "1": "y", "2": "z", "3": "w"}[a], vector_components_specified[1:]))

        # For vectors of length 2 and 4 using the even and odd syntax, use the
        # shorter .xyzw syntax.
        if vector_component_count == 2 or vector_component_count == 4:

            def sequential_fast_components():
                return "xyzw"[:vector_component_count]

            def even_fast_components():
                return sequential_fast_components()[::2]

            def odd_fast_components():
                return sequential_fast_components()[1::2]

            if vector_components_specified == "even":
                return even_fast_components()
            elif vector_components_specified == "odd":
                return odd_fast_components()

        # For vectors with more than 4 components matching specific patterns
        # are shortened to use the lo, hi, even, and odd syntax.
        # https://www.khronos.org/registry/cl/specs/opencl-1.1.pdf#page=167
        elif vector_component_count > 4 and vector_components_specified.startswith("s"):
            assert(vector_component_count >= 1)

            def sequential_components():
                return "0123456789abcdef"[0:vector_component_count]

            def lo_components():
                return sequential_components()[:vector_component_count // 2]

            def hi_components():
                return sequential_components()[vector_component_count // 2:]

            def even_components():
                return sequential_components()[::2]

            def odd_components():
                return sequential_components()[1::2]

            indices = vector_components_specified[1:].lower()
            if indices == lo_components():
                return "lo"
            elif indices == hi_components():
                return "hi"
            elif indices == even_components():
                return "even"
            elif indices == odd_components():
                return "odd"
        return vector_components_specified



from setuptools import setup, find_packages
try:
    import pypandoc
    long_description = pypandoc.convert("README.md", "rst")
except:
    from io import open
    long_description = open("README.md").read()


setup(
    name = "oclminify",
    description = "Minify OpenCL source files.",
    long_description = long_description,
    license = "BSD",
    version = "0.6.0",
    author = "StarByte Software, Inc.",
    author_email = "jbendig@starbytesoftware.com",
    maintainer = "James Bendig",
    url = "https://github.com/StarByteSoftware/oclminify",
    platforms = "Cross Platform",
    classifiers = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers ",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    packages = find_packages(),
    install_requires = [
        "pycparser>=2.14",
        "pycparserext>=2016.1",
    ],
    dependency_links = [
        # The 2.14 is over a year old as of 2016-07-15 and doesn't support
        # pragmas correctly. This commit is much newer and is known to work.
        "git+https://github.com/eliben/pycparser.git@ffd8cb7dfc4b80c79a500e27736db8f7bfc1186e#egg=pycparser-2.14",
    ],
    extras_require = {
        "build": ["PyOpenCL>=2016.1",],
    },
    entry_points = {
        "console_scripts": [
            "oclminify = oclminify.__main__:main",
        ],
    },
    test_suite = "tests.test_minifier",
)
